from math import e
from vcCommand import *
import os.path
import re
import vcMatrix, vcVector
import locale
locale.setlocale(locale.LC_NUMERIC,'C')


class TranslatorHelper:
  # Capsulates commonly used handles and post processor settings. Also caches translated lines.
  
  def __init__(self, app, program, uri):
    
    # 
    self.data =    ''         # Translated program lines
    self.positions = ''       # Translated robot positions
    self.position_names = []
    self.used_inputs = []
    self.used_outputs = []
    self.current_tool = ''
    self.nested = False
    self.nested_tool = ''
    self.inited_vars = {}

    # Helpers
    self.app = app
    self.uri = uri
    self.program = program
    
    self.app_version = 4.0
    try:
      self.app_version = float(app.ProductVersion[:app.ProductVersion.rfind('.')])
    except:
      pass
    
    # Properties
    self.command = getCommand()
    
    # Robot
    self.controller = program.Executor.Controller
    self.component = program.Executor.Component
    self.motiontarget = self.controller.createTarget()
    
    self.robot_joint_count = len([x for x in self.controller.Joints if not x.ExternalController])
    self.total_joint_count = len(self.controller.Joints)
    self.external_joint_count = self.total_joint_count - self.robot_joint_count
    
    self.indentation = '  '
    self.indent_depth = 0
    
    self.statement_translators = {
      VC_STATEMENT_BREAK:write_break,
      VC_STATEMENT_CALL:write_call,
      VC_STATEMENT_COMMENT:write_comment,
      VC_STATEMENT_CONTINUE:write_continue,
      VC_STATEMENT_DELAY:write_delay,
      VC_STATEMENT_HALT:write_halt,
      VC_STATEMENT_IF:write_if,
      VC_STATEMENT_LINMOTION:write_lin_motion,
      VC_STATEMENT_PATH:write_path,
      VC_STATEMENT_PRINT:write_print,
      VC_STATEMENT_PTPMOTION:write_ptp_motion,
      VC_STATEMENT_RETURN:write_return,
      VC_STATEMENT_SETBIN:write_set_bin,
      VC_STATEMENT_SETPROPERTY:write_set_property,
      VC_STATEMENT_SWITCHCASE:write_switch,
      VC_STATEMENT_WAITBIN:write_wait_bin,
      VC_STATEMENT_WHILE:write_while
      }


  def write_line(self, input):
    self.data += ('%s%s\n' % (self.current_indent(), input))
  
  def current_indent(self):
    return self.indentation * self.indent_depth
  
  def get_uri_folder(self):
    head, tail = os.path.split(self.uri)
    return head
  
  def get_job_name(self):
    head, tail = os.path.split(self.uri)
    name = tail[:tail.rfind('.')]
    return name
  

class RE(object):
  #Capsulates RE constants
  
  ANY = r'.+'
  ANYQ = r'.*'
  SPC = r'\s+'
  SPCQ = r'\s*'
  EOL = r'\n'
  EOS = r'$'
  INT = r'\s*-?\d+'
  NAME = r'\w+'
  NUM = r'\s*-?\d+\.?\d*[eE]?-?\d*'
  VAR = r'[a-z][a-z0-9]*(?![a-z\[\(])'
  ARRAY = r'[a-z][a-z0-9]*\[[0-9]+\]'
  FUNC = r'[a-z][a-z0-9]*\(.*?\)'
  STRING = r'\".*?\"'
  FLAGS = re.I|re.S


def postProcess(app, program, uri):
  # Entry point
  
  helper = TranslatorHelper(app, program, uri)
  write_header(helper)
  write_routine(helper, helper.program.MainRoutine)

  for routine in helper.program.Routines:
    write_routine(helper, routine)
  
  write_footer(helper)

  files = []
  with open(helper.uri, 'w') as file:
    file.write(helper.data)
    files.append(uri)
  return True, files


def write_header(helper):
  helper.write_line("define\n{\n\n}")


def write_footer(helper):
  helper.write_line("closestop\n{\n\n}\nerrorstop\n{\n\n}\n")


def write_routine(helper, routine):
  # Statements
  helper.write_line("main" if routine.Name == 'Main' else "void {}()".format(routine.Name))
  helper.write_line("{")
  helper.indent_depth += 1
  for statement in routine.Statements:
    if statement.IsEnabled:
      translator = helper.statement_translators.get(statement.Type, unknown)
      translator(statement, helper)

  helper.indent_depth -= 1
  helper.write_line("}")


def write_break(statement, helper):
  helper.write_line("break")


def write_call(statement, helper):
  helper.write_line("{}()".format(statement.Routine.Name))


def write_comment(statement, helper):
  helper.write_line("// {}".format(statement.Comment))


def write_continue(statement, helper):
  helper.write_line("continue")


def write_delay(statement, helper):
  pass


def write_halt(statement, helper):
  pass


def write_if(statement, helper):

  if not statement.IsEnabled:
    return
  # Currently does not work elegantly on nested nests
  helper.nested_tool = helper.current_tool
  helper.nested = True  

  # If Statements
  if statement.ThenScope.Statements:
    val = parse_value(statement.Condition)
    helper.write_line("if ({})".format(val))
    helper.write_line("{")    
    helper.indent_depth += 1  
    for stat in statement.ThenScope.Statements:
      if stat.IsEnabled:
        translator = helper.statement_translators.get(stat.Type, unknown)
        translator(stat, helper)
    helper.indent_depth -= 1
    helper.write_line("}")

  # ElseIf Statements
  for elseifs in statement.ElseIfScopes:
    val = parse_value(elseifs.Condition)
    helper.write_line("else if ({})".format(val))
    helper.write_line("{")
    helper.indent_depth += 1  
    for stat in elseifs.Statements:
      if stat.IsEnabled:
        translator = helper.statement_translators.get(stat.Type, unknown)
        translator(stat, helper)
    helper.indent_depth -= 1
    helper.write_line("}")

  # Else Statements
  if statement.ElseScope.Statements:
    helper.write_line("else")
    helper.write_line("{")    
    helper.indent_depth += 1  
    for stat in statement.ElseScope.Statements:
      if stat.IsEnabled:
        translator = helper.statement_translators.get(stat.Type, unknown)
        translator(stat, helper)
    helper.indent_depth -= 1
    helper.write_line("}")
  helper.nested_tool = ''  
  helper.nested = False


def write_lin_motion(statement, helper):

  statement.writeToTarget(helper.motiontarget)
  change_tcp(statement, helper)
  mtx = statement.Positions[0].PositionInReference
  to_list = [mtx.P.X , mtx.P.Y , mtx.P.Z , mtx.WPR.X , mtx.WPR.Y , mtx.WPR.Z] # type -> Real
  # From list to a pure string
  stringified = ', '.join(str(round(c, 2)) for c in to_list)
  pos_name = statement.Positions[0].Name
  # Triple curly brackets for escape string
  helper.write_line("float[] target{0} = {{{1}}}".format(
    pos_name,
    stringified)
    )
  helper.write_line("Line(\"CAR\", target{0}, {1}, {2}, {3}, false)".format(
    pos_name,
    int(statement.MaxSpeed),
    int((statement.MaxSpeed/statement.Acceleration)*1000),
    int(statement.AccuracyValue))
    )


def write_path(statement, helper):
  pass
  

def write_print(statement, helper):
  pass


def write_ptp_motion(statement, helper):

  statement.writeToTarget(helper.motiontarget)
  jv = helper.motiontarget.JointValues
  # From list to a pure string
  stringified = ', '.join(str(round(j, 2)) for j in jv)
  pos_name = statement.Positions[0].Name
  # Triple curly brackets for escape string
  helper.write_line("float[] target{0} = {{{1}}}".format(
    pos_name,
    stringified)
    )
  helper.write_line("PTP(\"JPP\", target{0}, {1}, 250, {2}, false)".format(
    pos_name,
    int(statement.JointSpeed * 100),
    int(statement.AccuracyValue))
    )  
 

def write_return(statement, helper):
  helper.write_line("return")


def write_set_bin(statement, helper):
  pass


def write_set_property(statement, helper):
  
  my_value = parse_value(statement.ValueExpression)
  helper.write_line("{0}{1} = {2}".format(
    get_prop_type(statement, helper),
    statement.TargetProperty, 
    my_value)
    )


def write_switch(statement, helper):

  if not statement.IsEnabled:
    return
  default_flag = False
  helper.write_line("switch({0})".format(statement.Condition))
  helper.write_line("{")
  helper.indent_depth += 1

  for s in statement.Cases:
    if s.CaseCondition == 'default':
      helper.write_line("default")
      default_flag = True
    else:
      helper.write_line("case {}".format(s.CaseCondition))

    for stat in s.Statements:
      helper.indent_depth += 1
      if stat.IsEnabled:
        translator = helper.statement_translators.get(stat.Type, unknown)
        translator(stat, helper)
      helper.indent_depth -= 1
    helper.write_line("\tbreak")

  if not default_flag:
    helper.write_line("default")
    helper.write_line("\tbreak")
  helper.indent_depth -= 1
  helper.write_line("}")


def write_wait_bin(statement, helper):
  pass


def write_while(statement, helper):

  val = parse_value(statement.Condition)
  helper.write_line("while ({})".format(val))
  helper.write_line("{")
  helper.indent_depth += 1
  
  for stat in statement.Scope.Statements:
    translator = helper.statement_translators.get(stat.Type, unknown)
    if stat.IsEnabled:
      translator(stat, helper)

  helper.indent_depth -= 1
  helper.write_line("}")


def unknown(statement, helper):
  print '> Unsupported statement type skipped:', statement.Type


def parse_value(val):

  if val == "True":
    return "true"
  elif val == "False":
     return "false"
  else:
    return val


def change_tcp(statement, helper):

  # Picking the correct tool for cartesian coordinates
  if statement.Tool:
    mtx = statement.Tool.PositionMatrix
    to_list = [mtx.P.X , mtx.P.Y , mtx.P.Z , mtx.WPR.X , mtx.WPR.Y , mtx.WPR.Z]
    stringified = ', '.join(str(round(t, 2)) for t in to_list)
    tool_name = statement.Tool.Name
    # Triple curly brackets for escape string
    if helper.nested:
      if helper.nested_tool != statement.Tool.Name:
        helper.nested_tool = statement.Tool.Name
        helper.write_line("float[] {0} = {{{1}}}".format(
          tool_name,
          stringified)
          )    
        helper.write_line("ChangeTCP({})".format(helper.nested_tool))
    else:
      if helper.current_tool != statement.Tool.Name:
        helper.write_line("float[] {0} = {{{1}}}".format(
          tool_name,
          stringified)
          )    
        helper.current_tool = statement.Tool.Name
        helper.write_line("ChangeTCP({})".format(helper.current_tool)) 
        
  else:
    if helper.nested:
      if helper.nested_tool != "NOTOOL":
        helper.nested_tool = "NOTOOL"
        helper.write_line("ChangeTCP(\"{}\")".format(helper.nested_tool))        
    else:
      if helper.current_tool != "NOTOOL":
        helper.current_tool = "NOTOOL"
        helper.write_line("ChangeTCP(\"{}\")".format(helper.current_tool))


def get_prop_type(statement, helper):
  # Loop through routine variables and check that the value exists, return the type
  # This could be a lot more elegant, problem is that statement is "SetProperty"
  check_keys(statement, helper)
  # Check for local variables and whether it has already been initalized
  for props in statement.ParentRoutine.Properties:
    if props.Type == VC_BOOLEAN and statement.TargetProperty == props.Name:
      return ("" if check_init_status(statement, helper) else "bool ")
    elif props.Type == VC_REAL and statement.TargetProperty == props.Name:
      return ("" if check_init_status(statement, helper) else "float ")
    elif props.Type == VC_INTEGER and statement.TargetProperty == props.Name:
      return ("" if check_init_status(statement, helper) else "int ")
    elif props.Type == VC_STRING and statement.TargetProperty == props.Name:
      return ("" if check_init_status(statement, helper) else "string ")
    else:
      continue


def check_keys(statement, helper):
  
  if statement.ParentRoutine.Name in helper.inited_vars:
    return
  else:
    helper.inited_vars[statement.ParentRoutine.Name] = []
 

def check_init_status(statement, helper):

  if statement.TargetProperty in helper.inited_vars[statement.ParentRoutine.Name]:
    return True
  else:
    helper.inited_vars[statement.ParentRoutine.Name].append(statement.TargetProperty)
    return False